(* Das Beispiel wurde dem Tutorial (C++) von www.snorre-dev.com entnommen und
   nach Delphi bersetzt
   J. Rathlev, Jan. 2002, info@rathlev-home.de *)

unit Lesson5;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Direct3D8, d3dx8, ExtCtrls;

const
  MPath = 'Media\';       // Pfad mit Texturen
  CubeCount = 36;
  
type
// Unsere Struktur, in der wir die Dreiecke speichern
  TMyVertex = record
    x,y,z     : single;   // Position des Vertex
    color     : dword;    // Farbe des Vertex
    tu,tv     : single;   // Die Koordinaten der Textur
    end;

  // Vertexliste fr Wrfel
  TCube = array [0..CubeCount-1] of TMyVertex;

const
// Beschreibung des Vertextyps: Mit D3DFVF_DIFFUSE sagen wir DX, das unsere
// Struktur eine Farbe hat. D3DFVF_XYZ bedeutet, das es sich um ein untransformiertes
// Vertex handelt
  D3D8T_CUSTOMVERTEX =D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1;

  NormCube : TCube = (
    (x :-1.0; y :-1.0; z :-1.0; color : $000000FF; tu : 0; tv : 1),  // Vorn
    (x :-1.0; y : 1.0; z :-1.0; color : $000000FF; tu : 0; tv : 0),
    (x : 1.0; y : 1.0; z :-1.0; color : $000000FF; tu : 1; tv : 0),
    (x : 1.0; y : 1.0; z :-1.0; color : $000000FF; tu : 1; tv : 0),
    (x : 1.0; y :-1.0; z :-1.0; color : $000000FF; tu : 1; tv : 1),
    (x :-1.0; y :-1.0; z :-1.0; color : $000000FF; tu : 0; tv : 1),

    (x :-1.0; y :-1.0; z : 1.0; color : $00FF00FF; tu : 0; tv : 1),  // Links
    (x :-1.0; y : 1.0; z : 1.0; color : $00FF00FF; tu : 0; tv : 0),
    (x :-1.0; y : 1.0; z :-1.0; color : $00FF00FF; tu : 1; tv : 0),
    (x :-1.0; y : 1.0; z :-1.0; color : $00FF00FF; tu : 1; tv : 0),
    (x :-1.0; y :-1.0; z :-1.0; color : $00FF00FF; tu : 1; tv : 1),
    (x :-1.0; y :-1.0; z : 1.0; color : $00FF00FF; tu : 0; tv : 1),

    (x : 1.0; y :-1.0; z : 1.0; color : $0000FF00; tu : 0; tv : 1),  // Hinten
    (x : 1.0; y : 1.0; z : 1.0; color : $0000FF00; tu : 0; tv : 0),
    (x :-1.0; y : 1.0; z : 1.0; color : $0000FF00; tu : 1; tv : 0),
    (x :-1.0; y : 1.0; z : 1.0; color : $0000FF00; tu : 1; tv : 0),
    (x :-1.0; y :-1.0; z : 1.0; color : $0000FF00; tu : 1; tv : 1),
    (x : 1.0; y :-1.0; z : 1.0; color : $0000FF00; tu : 0; tv : 1),

    (x : 1.0; y :-1.0; z :-1.0; color : $0000FFFF; tu : 0; tv : 1),  // Rechts
    (x : 1.0; y : 1.0; z :-1.0; color : $0000FFFF; tu : 0; tv : 0),
    (x : 1.0; y : 1.0; z : 1.0; color : $0000FFFF; tu : 1; tv : 0),
    (x : 1.0; y : 1.0; z : 1.0; color : $0000FFFF; tu : 1; tv : 0),
    (x : 1.0; y :-1.0; z : 1.0; color : $0000FFFF; tu : 1; tv : 1),
    (x : 1.0; y :-1.0; z :-1.0; color : $0000FFFF; tu : 0; tv : 1),

    (x :-1.0; y : 1.0; z :-1.0; color : $00FF0000; tu : 0; tv : 1),  // Oben
    (x :-1.0; y : 1.0; z : 1.0; color : $00FF0000; tu : 0; tv : 0),
    (x : 1.0; y : 1.0; z : 1.0; color : $00FF0000; tu : 1; tv : 0),
    (x : 1.0; y : 1.0; z : 1.0; color : $00FF0000; tu : 1; tv : 0),
    (x : 1.0; y : 1.0; z :-1.0; color : $00FF0000; tu : 1; tv : 1),
    (x :-1.0; y : 1.0; z :-1.0; color : $00FF0000; tu : 0; tv : 1),

    (x : 1.0; y :-1.0; z :-1.0; color : $00FFFF00; tu : 0; tv : 1),  // Unten
    (x : 1.0; y :-1.0; z : 1.0; color : $00FFFF00; tu : 0; tv : 0),
    (x :-1.0; y :-1.0; z : 1.0; color : $00FFFF00; tu : 1; tv : 0),
    (x :-1.0; y :-1.0; z : 1.0; color : $00FFFF00; tu : 1; tv : 0),
    (x :-1.0; y :-1.0; z :-1.0; color : $00FFFF00; tu : 1; tv : 1),
    (x : 1.0; y :-1.0; z :-1.0; color : $00FFFF00; tu : 0; tv : 1));

type
  TSample3DForm = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private-Deklarationen }
    // Das main Direct3D Interface, es wird zum initialisieren und schlieen von D3D bentigt
    lpd3d            : IDIRECT3D8;

    // Das D3DDevice wird zum Rendern benutzt und spiegelt den Bildschirm mit allen
    // Funktionen wieder. Wenn wir das D3DRender Interface erstellen, verndern oder 
    // das Bild rendern, so machen wir das ber dieses Interface
    lpd3ddevice      : IDirect3DDevice8;

    // Buffer, der die Vertizes des Wrfels enthlt
    CubeVB : IDirect3DVertexBuffer8;

    // Textur des Wrfels
    CubeTexture   : IDIRECT3DTEXTURE8;

    RotX,RotY,RotZ   : single  ;  //Rotation des Wrfels um die drei Achsen
    Animate          : boolean;

    procedure FatalError(hr : HResult; FehlerMsg : string);
    function D3DFind16BitMode : TD3DFORMAT;
    procedure D3DInit;
    procedure D3DShutdown;
    procedure D3DInitScene;
    procedure D3DKillScene;
    procedure D3DRender;
    procedure MyIdleHandler (Sender: TObject; var Done: Boolean);
  public
    { Public-Deklarationen }
  end;

var
  Sample3DForm: TSample3DForm;

implementation

{$R *.DFM}

// Initialisieren aller Variablen
procedure TSample3DForm.FormCreate(Sender: TObject);
begin
  lpd3d:=nil;
  lpd3ddevice:=nil;
  CubeVB:=nil;
  CubeTexture:=nil;
  //Initialisiere die Rotation 
  RotX:=0;
  RotY:=0;
  RotZ:=0;
  Animate:=false;
  Application.OnIdle:= MyIdleHandler;
  end;

// Fataler Fehler. Meldung und Programmende
procedure TSample3DForm.FatalError(hr : HResult; FehlerMsg : string);
var
  s : string;
begin
  if hr<>0 then s:=D3DXErrorString(hr)+#13+FehlerMsg else s:=FehlerMsg;
  D3DKillScene;
  D3DShutdown;
  MessageDlg(s,mtError,[mbOK],0);
  close;
  end;

function TSample3DForm.D3DFind16BitMode : TD3DFORMAT;
var
  hr    : HRESULT;
begin
  hr:=lpd3d.CheckDeviceType (D3DADAPTER_DEFAULT,
                             D3DDEVTYPE_HAL,
                             D3DFMT_R5G6B5,      // Format Primary
                             D3DFMT_R5G6B5,      // Format Buffer
                             FALSE);
  if (SUCCEEDED(hr)) then begin
    Result:=D3DFMT_R5G6B5; exit;
    end;

  hr:=lpd3d.CheckDeviceType (D3DADAPTER_DEFAULT,
                             D3DDEVTYPE_HAL,
                             D3DFMT_X1R5G5B5,
                             D3DFMT_X1R5G5B5,
                             FALSE);
  if (SUCCEEDED(hr)) then begin
    Result:=D3DFMT_X1R5G5B5; exit;
    end;
  FatalError(0,'Der erforderliche D3D Modus wird von Ihrer Karte nicht untersttzt');
  Result:=0;
  end;

// Mit dieser Funktion initialisieren wir D3D
procedure TSample3DForm.D3DInit;
var
  hr    : HRESULT;
  d3dpp : TD3DPRESENTPARAMETERS;
begin
  //Erstelle Direct3D! Mu immer als erstes erstellt werden
  //Immer D3D_SDK_VERSION als Version setzen
  lpd3d:=Direct3DCreate8(D3D_SDK_VERSION);
  if(lpd3d=nil) then FatalError(0,'Fehler beim Erstellen von Direct3D!');

  // Setze D3DPRESENT_PARAMETERS auf 0, sonst knnten wir probleme mit lteren
  // Eintragungen bzw. unkontrollierbaren Ergebnissen bekommen!
  // Sollten wir bei allen DirectX Strukturen machen
  ZeroMemory(@d3dpp,sizeof(d3dpp));

  // Hiermit werden alte Frames gelscht, denn wir brauchen sie nicht
  with d3dpp do begin
    SwapEffect:=D3DSWAPEFFECT_DISCARD;
    hDeviceWindow:=Handle;      // Dies ist unser HWND von TForm
    BackBufferCount:=1;         // 1 Backbuffer

    // Wir brauche einen Z-Buffer also schalten wir ihn ein
    EnableAutoDepthStencil := TRUE;
    AutoDepthStencilFormat := D3DFMT_D16;

    Windowed          := FALSE;
    BackBufferWidth   := 640;
    BackBufferHeight  := 480;
    BackBufferFormat  := D3DFind16BitMode;
    end;

  //Nachdem wir die D3DPRESENT_PARAMETERS Struktur ausgefllt haben, sind wir
  // endlich so weit unser D3D Device zu erstellen
  hr:=lpd3d.CreateDevice(D3DADAPTER_DEFAULT,
                         D3DDEVTYPE_HAL,
                         Handle,
                         D3DCREATE_SOFTWARE_VERTEXPROCESSING,
//                         D3DCREATE_HARDWARE_VERTEXPROCESSING,
                         d3dpp,
                         lpd3ddevice);
  if FAILED(hr) then FatalError(hr,'Fehler beim Erzeugen des 3D-Device');
  end;

// *** D3DShutdown hier werden die Resourcen von D3D wieder freigegeben
procedure TSample3DForm.D3DShutdown;
begin
  if assigned(lpd3ddevice) then lpd3ddevice:=nil;
  if assigned(lpd3d) then lpd3d:=nil;
  end;

procedure TSample3DForm.D3DInitScene;
var
  hr         : HRESULT;
  vbVertices : pByte;
  ViewMatrix,
  matProj    : TD3DXMATRIX;
begin
  // Hier wird der Vertex Buffer erstellt, der gro genug ist um alle Vertizes zu enthalten.
  if assigned(lpd3ddevice) then with lpd3ddevice do begin
    hr:=CreateVertexBuffer (sizeof(TCube),
                            D3DUSAGE_WRITEONLY, // Nur Schreibzugriffe
                            D3D8T_CUSTOMVERTEX, // Unser Vertex
                            D3DPOOL_MANAGED,
                            CubeVB);              // Pointer zu unserem Buffer

    if FAILED(hr) then FatalError(0,'Fehler beim Erstellen des Vertex Buffers');

    // Nun kopieren wir unsere Vertizes in den Buffer
    // Wir mssen es zuvor mit Lock festhalten, um es bearbeiten zu knnen
    with CubeVB do begin
      hr:=Lock(0, // Offset, an dem wir beginnen
               0, // Gre des locks ( 0 fr alles )
               vbVertices, // Wenn erfolgreich dann hier ablegen
               0); // sonstige Flags
      if FAILED(hr) then FatalError(0,'Fehler beim Locken des Vertex-Buffers');
      // Hier wird der Vertexbuffer kopiert.
      Move(NormCube,vbVertices^,SizeOf(TCube));
      // Und wieder loslassen
      Unlock;
      end;

    SetRenderState(D3DRS_CULLMODE,D3DCULL_CCW);
    SetRenderState(D3DRS_LIGHTING,0);
    SetRenderState(D3DRS_ZENABLE,0);

    // Hier erstellen wir unseren SichtMatrix. Denkt einfach es ist
    // eure Kamera, von der aus wir sehen. Als erstes setzen wir die Kamera
    // um 8 Einheiten zurck auf der Z-Achse.
    D3DXMatrixLookAtLH (ViewMatrix,D3DXVECTOR3(0.0,0.0,-8.0),
                                   D3DXVECTOR3(0.0,0.0,0.0),
                                   D3DXVECTOR3(0.0,1.0,0.0));

    // Da sich unsere Camera nicht bewegt legen wir sie einfach fest
    SetTransform(D3DTS_VIEW,ViewMatrix);

    D3DXMatrixPerspectiveFovLH(matProj, //Resultierende Matrix
                               D3DX_PI/4,//Radius der Ansicht
                               640/480,
                               1.0,     // Mindeste Nhe
                               100.0);  // Maximal sichtbare Entfernung

    // Unsere Projektion wird sich niemals bewegen, also setzen wir sie fest
    SetTransform(D3DTS_PROJECTION,matProj );

    // Mit D3DTSS_COLLORP wird festgelegt, wie die Farbe jedes einzelnen Pixels
    // verarbeitet wird. D3DTOP_SELECTARG1 verweist auf D3DTSS_COLORARG1
    SetTextureStageState(0,D3DTSS_COLOROP, D3DTOP_SELECTARG1);
    // Mit D3DTSS_COLORARG1 wird festgelegt, da die Farbe nur von der Textur
    // genommen wird und von nichts anderem.
    SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE);
    // Wir benutzen kein Alpha blending, also schalten wir es ab. Dadurch wird
    // der Code etwas schneller
    SetTextureStageState(0,D3DTSS_ALPHAOP,   D3DTOP_DISABLE);
    // MAGFILTER ist dafr, wenn es kleiner ist als unser Objekt.
    // MINFILTER ist genau das umgekehrte
    // Standard ist D3DTEXF_LINEAR. Dieses ist zwar langsamer als D3DTEXF_POINT,
    // sieht aber besser aus.
    SetTextureStageState(0,D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
    SetTextureStageState(0,D3DTSS_MINFILTER, D3DTEXF_LINEAR);

    // Ich glaube diese Funktion bruchte ich nicht zu beschreiben, aber hier
    // wird die Textur geladen.
    hr:=D3DXCreateTextureFromFile(lpd3ddevice,pchar(MPath+'ieap-kiste-g.bmp'),CubeTexture);
    if(FAILED(hr)) then FatalError(0,'Fehler beim Laden der Textur');
    end;
  Animate:=true;
  end;

procedure TSample3DForm.D3DKillScene;
begin
  CubeVB:=nil;
  end;

procedure TSample3DForm.D3DRender;
var
  WorldMatrix,TempMatrix  : TD3DXMATRIX;
begin
  RotY:=RotY+0.03;
  RotX:=RotX+0.04;
  RotZ:=RotZ+0.02;

  if assigned(lpd3ddevice) then with lpd3ddevice do begin
    Clear(0,           // Wieviel Rechtecke lschen? 0 Lscht alle
          nil,         // Pointer zu den Rechtecken. nil = Ganzer Bildschirm
          D3DCLEAR_TARGET,
          D3DCOLOR_XRGB(0,0,0), //Hintergrundfarbe schwarz
          1,           // Lsche ZBuffer ( Wir haben momentan noch keinen )
          0 );

    if SUCCEEDED(BeginScene) then begin
      // Hier wird unsere Textur geladen
      SetTexture(0,CubeTexture);

      // Vertex Shader sind wirklich komplex, aber es lassen sich damit gute Effekte
      // erzielen. Genauere Beschreibungen in der SDK, denn alles hier niederschreiben
      // sprengt den Rahmen eines Tutorials
      SetVertexShader(D3D8T_CUSTOMVERTEX);

      // Die D3D Renderfunktionen lesen aus Streams. Hier sagen wir DX welchen Stream
      // es verwenden soll
      SetStreamSource(0,CubeVB,sizeof(TMyVertex));

      // Die Rotation um alle Achsen.
      D3DXMatrixRotationYawPitchRoll(TempMatrix,RotY,RotX,RotZ);
      // Verschiebe den Wrfel etwas nach vorn
      D3DXMatrixTranslation(WorldMatrix,0.0,0.0,-2.0);
      // Berechne die Welt-Matrix fr die Transformation
      D3DXMatrixMultiply(WorldMatrix,TempMatrix,WorldMatrix);
      SetTransform(D3DTS_WORLD,WorldMatrix);

      // Zeichnen des Wrfels
      DrawPrimitive(D3DPT_TRIANGLELIST,0,12);

      EndScene;
      end;

   // Zeige Resultate auf dem Bildschirm
    Present(nil,nil,0,nil);
    end;
  end;

procedure TSample3DForm.FormShow(Sender: TObject);
begin
  D3DInit;     // Initialisieren von D3D
  D3DInitScene;
  D3DRender;   // Zeichne unsere Grafiken
  end;

procedure TSample3DForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  // Lsche die D3D Scene bevor wir D3D beenden
  D3DKillScene;
  // Lsche D3D
  D3DShutdown;
  end;

procedure TSample3DForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_ESCAPE then close;
  if Key=VK_SPACE then Animate:=not Animate;
  end;

procedure TSample3DForm.MyIdleHandler (Sender: TObject; var Done: Boolean);
begin
  if Animate then D3DRender;
  Done:=false;
  end;

end.
